# Home
## Introduction
AzCam is a software framework for the acquisition and analysis of image data from scientific imaging systems as well as control of instrumentation. It is intended to be customized for specific hardware, observational, and data reduction needs.

AzCam is based on the concept of *tools* which are the interfaces to both hardware and software code.  Examples of tools are *instrument* which controls instrument hardware, *telescope* which interfaces to a telescope, and *linearity* which acquires and analyzes a sensor for linearity, and *exposure* tool which controls a scientific observation by interfacing with many different tools. As an example, the *exposure* tool might need to move a telescope and multiple filter wheels, control a shutter, operate a camera, display the resultant image, and begin data reduction of that image.  

AzCam is not appropriate for consumer-level cameras and is not intended to have a common API across all environments.

## Extension Packages
AzCam *extension packages* define resources which extend AzCam's functionality. Examples include:

  * [azcam-arc](https://github.com/mplesser/azcam-arc) adds support for ARC cameras
  * [azcam-archon](https://github.com/mplesser/azcam-archon) adds support for STA Archon cameras
  * [azcam-ds9](https://github.com/mplesser/azcam-ds9) adds support for the SAO ds9 image display
  * [azcam-mag](https://github.com/mplesser/azcam-mag) adds support for Magellan/ITL cameras
  * [azcam-imageserver](https://github.com/mplesser/azcam-imageserver) adds support for remote image servers
  * [azcam-cryocon](https://github.com/mplesser/azcam-cryocon) adds support for CryoCon temperature controllers
  * [azcam-qhy](https://github.com/mplesser/azcam-qhy) adds support for QHYCCD CMOS cameras
  * [azcam-testers](https://github.com/mplesser/azcam-testers) adds support for image sensor testing such as QE, CTE, PRNU, etc.
  * [azcam-flaskserver](https://github.com/mplesser/azcam-flaskserver) adds support for a flask-based web server for web client apps
  * [azcam-scripts](https://github.com/mplesser/azcam-scripts) adds a collection of general purpose scripts
  * [azcam-focus](https://github.com/mplesser/azcam-focus), adds routines to acquire focus images
  * [azcam-observe](https://github.com/mplesser/azcam-observe) add observing scripts whcih support a Qt-basewd GUI and command line interface

  * [azcam-queue](https://github.com/mplesser/azcam-queue) adds server support for web-enabled observing queue similar to observing scripts 

Some packages act as *environments* to define code and data files used for specific hardware systems. Examples include:

  * [azcam-bok](https://github.com/mplesser/azcam-bok) for the UArizona Bok telescope environment
  * [azcam-mont4k](https://github.com/mplesser/azcam-mont4k) for the UArizona Mont4k instrument environment
  * [azcam-vatt](https://github.com/mplesser/azcam-vatt) for the VATT VattSpec and Vatt4k cameras

## Applications
AzCam *applications* are stand-alone programs which utilize AzCam functionality. The most important application is *azcamserver* which usually defines to tools which interface to all hardware. Most other applications are clients which may connect to an *azcamserver* application. There are many clients written in a variety of languages. Examples include:

  * [azcam-tool](https://github.com/mplesser/azcam-tool), a GUI using National Instruments LabVIEW
  * [azcam-webtools](https://github.com/mplesser/azcam-status), a web-based tool for status and exposure control

## Operation

There are three main operational modes of AzCam:

  1. A server, usually called *azcamserver*, which communicates directly or indirectly with system hardware.

  2. A console, usually called *azcamconsole*, typically implemented as an IPython command window that communicates with *azcamserver* over the socket interface.  It is used to acquire and analyze image data through the command line and python code.

  3. Applications, as described above. Client apps communicate with *azcamserver* over sockets or the web API.

While there are various pythonic ways to access the tools in code, they are always accessible from the *database* object `db`, which can always be accessed as `azcam.db`. For example, when defined, the `controller` tool can be accessed as `db.controller` and the `qe` tool can be accessed as `db.qe`.  In most environments these tool are mapped directly into the command line namespace, so in practice commands are executed directly as `object.method`, e.g. `exposure.expose(2.5, "dark", "dark image")`. 

In *azcamconsole*, there are *console tools* which allows communication from the console to an *azcamserver* application over the socket-based command server interface.  These tools typically expose only a limited set of methods to the standard server-side tools. So while `exposure.reset` may be available the command `exposure.set_video_gain` may not be.  These less commonly used commands are still accessible, but only with lower level code such as `server.command("controller.set_video_gain 2")`.

As an example, the code below can be used to set the current system wavelength and take an exposure.  It is assumed the the *azcam_itl* package has been installed.

```python
# server-side (azcamserver)
import azcam
import azcam_itl.server_itl
instrument, exposure = azcam.get_tools(["instrument", "exposure"])
instrument.set_wavelength(450)
wavelength = instrument.get_wavelength()
print(f"Current wavelength is {wavelength}")
exposure.expose(2., 'flat', "a 450 nm flat field image")

or

# client-side (azcamconsole)
import azcam
import azam_itl.console_itl
instrument, exposure = azcam.get_tools(["instrument", "exposure"])
instrument.set_wavelength(450)
wavelength = instrument.get_wavelength()
print(f"Current wavelength is {wavelength}")
exposure.expose(2., 'flat', "a 450 nm flat field image")
```

Both *azcamserver* and *azcamconsole* may also be called in a manner similar to:

```python
ipython.exe -i -m azcam_itl.server_itl -- -system DESI
ipython.exe -i -m azcam_itl.console_itl -- -system DESI
```

Example configuration code may be found in the various extension packages with names like `server_xxx.py` and `console_xxx.py` where `xxx` is the system name.

When working in a command line environment, it is often convenient to import commonly used commands into the CLI namespace. This provides direct access to objects and tools such as *db*, *exposure*, *controller*, and various pre-defined shortcuts. To do this, after configuring the environment, execute the command:

```python
from azcam.cli import *
```

And then the code above could be executed as:
```python
instrument.set_wavelength(450)
exposure.expose(2., 'flat', "a 450 nm flat field image")
```

## Help
AzCam is often used with IPython.  Help is then avaialble by typing `?xxx`, `xxx?`, `xxx??` or `help(xxx)` where `xxx` is an AzCam class, command, or object instance.

Useful links include:
* IPython <https://ipython.org>
* Python programming language <https://www.python.org>

## Command Structure
The AzCam command structure provides a fairly uniform interface which can be used from the local command line (CLI), a remote socket connection, or the web interface.  An example for taking a 2.5 second "flat field" exposure is:

Local CLI or script example:

```python
exposure.expose(2.5, 'flat', 'an image title')
```

A remote socket connection example:

```python
exposure.expose 2.5 flat "an image title"
```

Web (http) connection example:

```html
http://hostname:2403/api/exposure/expose?exposure_time=2.5&image_type=flat&image_title=an+image+title
```
Web support requires an extension package such as *azcam-flaskserver*.  Example URL's are:

```html
http://hostname:2403/status
```
```html
http://hostname:2403/exptool
```

The value *2403* here is the web server port configured by an environment.

## Shortcuts
When using IPython, the auto parenthesis mode allows typing commands without 
requiring the normal python syntax of ``command(par1, par2, ...)``. The equivalent 
shortcut/alias syntax is ``command par1 par2``. With IPython in this mode all commands may 
use this syntax.  There are a few special command line commands optionally defined within *azcamconsole* which 
are listed below. Most aliases are implemeted from a command such as `from shortcuts_console import *`.

  * *sav* - save the current parameters to the parameter file
  * *p* - toggle the command line printout of client commands and responses
  * *Run* - run a command found on the python search path, usually for scripts. Note the upper case *R* to distinguish it from IPython's built-in *run* magic command.
  * *gf* - try and go to current image folder.
  * *sf* - try and set the image folder to the current directory.
  * *bf* - browse for a file or folder usign a Tcl/Tk GUI.

## Code Descriptions
The links below describe some of the many classes and commands found in AzCam. Availability depends on configuration.

- [Tools](tools.md)
- [Classes](classes.md)
- [Commands](commands.md)

## Scripts
Scripts are python code modules which generally contain one function and are available from the command line or by importing a module or package such as the *azcam-scripts* package. They may be loaded automatically during enviroment configuration and are often called by the `Run` shortcut command. 

## Configuration Folders
There are two important folders which are usually defined by an environment for convenience:

  * *systemfolder* - the main folder where configuration code is located
  * *datafolder* - the root folder where data and parameters are saved, write access is required

## Programming

The link below contains information about AzCam programming and more advanced features.

- [Programming](programming.md)
